package swtpl.marsrover.gui;

import javax.swing.SwingUtilities;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JScrollPane;

import swtpl.marsrover.map.MapObject;
import swtpl.marsrover.map.MapStateEnum;
import swtpl.marsrover.net.Connection;
import swtpl.marsrover.net.InitHandler;
import swtpl.marsrover.net.MartianEvent;
import swtpl.marsrover.net.MartianHandler;
import swtpl.marsrover.net.ObstacleEvent;
import swtpl.marsrover.net.ObstacleHandler;
import swtpl.marsrover.net.RunContext;
import swtpl.marsrover.net.VehicleEvent;
import swtpl.marsrover.net.VehicleHandler;
import swtpl.marsrover.planner.GoalHandler;
import swtpl.marsrover.planner.JohoPlanner;
import swtpl.marsrover.planner.PathPlanner;
import swtpl.marsrover.planner.Planner;
import swtpl.marsrover.planner.RasterPlanner;

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.io.IOException;
import java.net.UnknownHostException;
import java.util.ArrayList;
import java.util.Vector;

public class ClientGui extends JPanel implements Runnable, InitHandler, VehicleHandler, ObstacleHandler, MartianHandler, PaintObserver {
	
    //all coordinates and distances are rover coordinates, not java coordinates
	
	//map
	private int mapHeight;
	private int mapWidth;
    private static final int SCALE = 400;
	
	//rover
	private int roverX;
    private int roverY;
    private int roverDir;
    private int roverVisionMin;
    private int roverVisionMax;
    private boolean isInitialized = false;
    private boolean isRoverInitialized = false;

    //home base
	private int baseX = 0;
	private int baseY = 0;

    //objects on the map
    private ArrayList<MapObject> craterList = new ArrayList<MapObject>();
    private ArrayList<MapObject> boulderList = new ArrayList<MapObject>();
    private ArrayList<Point> martianList = new ArrayList<Point>();
    
    //map colors
    private static final Color ROVER_VISION_COLOR = new Color(255, 255, 0);
    private static final Color BASE_COLOR = new Color(0, 250, 0);
    private static final Color ROVER_COLOR = new Color(83, 25, 147);
    private static final Color ROVER_VISION_DIR_COLOR = new Color(83, 25, 0);
    private static final Color CRATER_COLOR = new Color(147, 82, 0);
    private static final Color BOULDER_COLOR = new Color(120, 120, 120);
    private static final Color MARTIAN_COLOR = new Color(255, 37, 0);

    private Connection connection;
    private Vector<Painter> painter = new Vector<Painter>();
    private Graphics g;
    
	//start gui
    public static void main(String[] args) {
    	if( args.length != 2 ) {
    		System.err.println("USAGE: ClientGui <hostname> <port>");
    		System.exit(-1);
    	}
    	
    	new ClientGui(args[0], Integer.parseInt(args[1]));
    }

    public ClientGui(String hostname, int port){
    	try {
			connection = new Connection(hostname, port);
			connection.registerInitHandler(this);

			final Planner planner = new JohoPlanner(connection,this);
			
			connection.startReading(connection.getInputStream());
    	} catch (UnknownHostException e) {
			e.printStackTrace(System.err);
		} catch (IOException e) {
			e.printStackTrace(System.err);
		}
    }

	public void initialized(RunContext rc) {
		if(isInitialized){
			return;
		}
		isInitialized = true;
		
    	this.mapHeight = rc.getCardSizeY();
    	this.mapWidth = rc.getCardSizeX();
        this.roverVisionMin = rc.getMinSensorMM();
        this.roverVisionMax = rc.getMaxSensorMM();

		connection.registerMartianHandler(this);
		connection.registerObstacleHandler(this);
		connection.registerVehicleHandler(this);

        //init gui
    	SwingUtilities.invokeLater(this);		
	}

    //start gui
	public void run() {
		//create the window
        JFrame f = new JFrame("Map " + this.mapHeight + "x" + this.mapWidth);
        f.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
        JScrollPane jScrollPane = new JScrollPane(this);
        jScrollPane.setPreferredSize(new Dimension(Math.min(1000, this.mapWidth / SCALE + 20), Math.min(700, this.mapHeight / SCALE +20)));
        f.add(jScrollPane);
        f.pack();
        f.setVisible(true);
        
        //handle keyboard commands
        f.addKeyListener(new KeyAdapter() {
        	public void keyPressed(KeyEvent e) {
    			switch( e.getKeyCode() ) {
    			case KeyEvent.VK_LEFT:
    			case KeyEvent.VK_KP_LEFT:
    				connection.sendCommand(Connection.AccelerationCommand.NONE,Connection.TurningCommand.LEFT);
    				break;
    			case KeyEvent.VK_RIGHT:
    			case KeyEvent.VK_KP_RIGHT:
    				connection.sendCommand(Connection.AccelerationCommand.NONE,Connection.TurningCommand.RIGHT);
    				break;
    			case KeyEvent.VK_UP:
    			case KeyEvent.VK_KP_UP:
    				connection.sendCommand(Connection.AccelerationCommand.ACCELERATE,Connection.TurningCommand.NONE);
    				break;
    			case KeyEvent.VK_DOWN:
    			case KeyEvent.VK_KP_DOWN:
    				connection.sendCommand(Connection.AccelerationCommand.BREAK,Connection.TurningCommand.NONE);
    				break;
    			}
        	}
        });
    }

    //does the actual painting
    public void paintComponent(Graphics g) {
        super.paintComponent(g);
        if(!this.isRoverInitialized){
        	return;
        }

        this.g = g;
        //Draw rover view, always draw this first
        Graphics2D g2 = (Graphics2D)g;
        g2.rotate(-roverDir*2*Math.PI/3600, toJavaX(roverX), toJavaY(roverY));
        g.setColor(ClientGui.ROVER_VISION_COLOR);
        double ellipseHalfHeight = Math.sqrt(roverVisionMin*roverVisionMax);
        g2.fillOval(toJavaX(roverX - roverVisionMin), toJavaY(roverY + (int)ellipseHalfHeight), (roverVisionMin + roverVisionMax)/SCALE, (int)(2*ellipseHalfHeight/SCALE));
        g.setColor(ClientGui.ROVER_VISION_DIR_COLOR);
        g2.drawLine(toJavaX(roverX), toJavaY(roverY), toJavaX(roverX + roverVisionMax), toJavaY(roverY));
        g2.rotate(roverDir*2*Math.PI/3600, toJavaX(roverX), toJavaY(roverY));
        
        //Draw the home base, 5m, centered
        g.setColor(ClientGui.BASE_COLOR);
        fillCircle(this.baseX, this.baseY, 5000);

        //Draw the craters
        g.setColor(ClientGui.CRATER_COLOR);
        for(MapObject crater : this.craterList){
            fillCircle(crater.getX(), crater.getY(), crater.getRadius());
        }

        //Draw the boulders
        g.setColor(ClientGui.BOULDER_COLOR);
        for(MapObject boulder : this.boulderList){
            fillCircle(boulder.getX(), boulder.getY(), boulder.getRadius());
        }

        //Draw the rover, 0.5m
        g.setColor(ClientGui.ROVER_COLOR);
        fillCircle(roverX, roverY, 500);
        
        //Draw the martians, 0.4m
        g.setColor(ClientGui.MARTIAN_COLOR);
        for(Point martian : this.martianList){
        	fillCircle(martian.x, martian.y, 400);
        }
        
        for (Painter p : this.painter) {
        	p.paintComponent(this);
        }

        //focus on rover
        this.scrollRectToVisible(new Rectangle(toJavaX(roverX-roverVisionMax), toJavaY(roverY+roverVisionMax),
        			roverVisionMax*2/SCALE, roverVisionMax*2/SCALE));
    }

    //helpers for the painting
    private int toJavaX(int mapX){
    	return (this.mapWidth/2 + mapX) / SCALE;
    }
    private int toJavaY(int mapY){
    	return (this.mapHeight/2 - mapY) / SCALE;
    }
    public void fillCircle(int x, int y, int radius) {
    	g.fillOval(toJavaX(x-radius), toJavaY(y+radius), 2*radius/SCALE, 2*radius/SCALE);
    }
    public void drawCircle(int x, int y, int radius) {
    	g.drawOval(toJavaX(x-radius), toJavaY(y+radius), 2*radius/SCALE, 2*radius/SCALE);
    }
    public void drawRectangle(int x, int y, int width, int height) {
    	g.drawRect(toJavaX(x), toJavaY(y), Math.max(1,width/SCALE), Math.max(1,height/SCALE));
    }
    public void drawLine(int x1, int y1, int x2, int y2) {
    	g.drawLine(toJavaX(x1), toJavaY(y1), toJavaX(x2), toJavaY(y2));
    }

    //map update handlers
	public void changed(VehicleEvent ve) {
		this.isRoverInitialized = true;
		martianList.clear();
		roverX = ve.getPosX();
		roverY = ve.getPosY();
		roverDir = ve.getDir();
		repaint();
	}
	public void recognized(ObstacleEvent oe) {
		MapObject mo = new MapObject(oe.getX(),oe.getY(),oe.getRadius());
		if( oe.getSort() == MapStateEnum.BOULDER )
			boulderList.add(mo);
		else
			craterList.add(mo);
	}
	public void recognized(MartianEvent me) {
		Point pt = new Point(me.getX(),me.getY());
		martianList.add(pt);
	}

	//set size of the JLabel
    public Dimension getPreferredSize() {
        return new Dimension(this.mapHeight/SCALE, this.mapWidth/SCALE);
    }
    
    public void addPainter(Painter p) {
    	this.painter.add(p);
    }

	@Override
	public void setColor(Color c) {
		this.g.setColor(c);
	}
}