package swtpl.marsrover.net;

import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.Socket;
import java.net.UnknownHostException;
import java.util.HashSet;
import java.util.Set;

import swtpl.marsrover.map.MapStateEnum;

public class Connection {
	public enum AccelerationCommand {
		NONE,
		ACCELERATE,
		BREAK
	}
	public enum TurningCommand {
		NONE,
		LEFT,
		RIGHT
	}
	private class Obstacle {
		int mx;
		int my;
		int mr;
		public Obstacle(int x,int y,int r) {
			mx = x;
			my = y;
			mr = r;
		}
		public int hashCode() {
			return mx * 21474851 + my * 10656727 + mr;
		}
		public boolean equals(Object o) {
			if( o instanceof Obstacle ) {
				Obstacle obs = (Obstacle)o;
				return (mx == obs.mx) && (my == obs.my) && (mr == obs.mr); 
			}
			return false;
		}
	}
	private Socket msock;
	private	OutputStream msockstream;
	private Set<VehicleHandler> mvehicles;
	private Set<ObstacleHandler> mobstacles;
	private Set<MartianHandler> mmartians;
	private Set<CrashHandler> mcrashs;
	private Set<CleanupBase> mcleanups;
	private Set<InitHandler> minits;
	private Set<Obstacle> mdetobstacles;
	public Connection(String hostname,int port) throws UnknownHostException, IOException {
		msock = new Socket(hostname,port);
		msock.setTcpNoDelay(true);
		msockstream = msock.getOutputStream();
		mvehicles = new HashSet<VehicleHandler>(3);
		mobstacles = new HashSet<ObstacleHandler>(3);
		mmartians = new HashSet<MartianHandler>(3);
		mcrashs = new HashSet<CrashHandler>(3);
		mcleanups = new HashSet<CleanupBase>(3);
		minits = new HashSet<InitHandler>(3);
		mdetobstacles = new HashSet<Obstacle>();
	}
	
	public InputStream getInputStream() throws IOException {
		return msock.getInputStream();
	}
	
	public OutputStream getOutputStream() throws IOException {
		return msock.getOutputStream();
	}
	
	public void startReading(final InputStream in) {
		new Thread(new Runnable() {

			@Override
			public void run() {
				try {
					int input;
					// TODO: check capacity
					StringBuilder sb = new StringBuilder(500);
					while( (input = in.read()) != -1 ) {
						if( input == ';' ) {
							parseAndNotify(sb);
							sb = new StringBuilder(500);
						} else if( input != '.' )
							sb.append((char)input);
					}
					// Ends thread.
				} catch( IOException ignored ) {}
				
				System.exit(0);
			}
			
		}).start();
	}
	
	public void registerVehicleHandler(VehicleHandler vh) {
		synchronized( mvehicles ) {
			mvehicles.add(vh);
		}
	}
	
	public void registerObstacleHandler(ObstacleHandler oh) {
		synchronized( mobstacles ) {
			mobstacles.add(oh);
		}
	}
	
	public void registerMartianHandler(MartianHandler mh) {
		synchronized( mmartians ) {
			mmartians.add(mh);
		}
	}
	
	public void registerCrashHandler(CrashHandler ch) {
		synchronized( mcrashs ) {
			mcrashs.add(ch);
		}
	}
	
	public void registerCleanup(CleanupBase cb) {
		synchronized( mcleanups ) {
			mcleanups.add(cb);
		}
	}
	
	public void registerInitHandler(InitHandler ih) {
		synchronized( minits ) {
			minits.add(ih);
		}
	}
	
	private final void parseAndNotify(StringBuilder sb) {
		char msgtype = sb.charAt(0);
		String[] split = sb.substring(2).split(" ");
		switch(msgtype) {
		case 'I':
			int dxmm = Integer.parseInt(split[0]);
			int dymm = Integer.parseInt(split[1]);
			int timelimit = Integer.parseInt(split[2]);
			int minsensormm = Integer.parseInt(split[3]);
			int maxsensormm = Integer.parseInt(split[4]);
			int maxspeedmmps = Integer.parseInt(split[5]);
			int tenthdegps = Integer.parseInt(split[6]);
			int tenthdegpshard = Integer.parseInt(split[7]);
			RunContext rc = new RunContext(dxmm,dymm,timelimit,minsensormm,maxsensormm,maxspeedmmps,tenthdegps,tenthdegpshard);
			synchronized( minits ) {
				for( InitHandler ih : minits )
					ih.initialized(rc);
			}
			/*System.out.print("Read initial message: dx = ");
			System.out.print(dxmm);
			System.out.print("; dy = ");
			System.out.print(dymm);
			System.out.print("; timelimit = ");
			System.out.print(timelimit);
			System.out.print("; min-sensor = ");
			System.out.print(minsensormm);
			System.out.print("; max-sensor = ");
			System.out.print(maxsensormm);
			System.out.print("; max-speed = ");
			System.out.print(maxspeedmmps);
			System.out.print("; max-turn = ");
			System.out.print(tenthdegps);
			System.out.print("; max-hard-turn = ");
			System.out.println(tenthdegpshard);
			System.out.flush();*/
			break;
		case 'T':
			int timestamp = Integer.parseInt(split[0]);
			VehicleEvent.VehicleControlAcceleration vca;
			switch( split[1].charAt(0) ) {
			case 'a':
				vca = VehicleEvent.VehicleControlAcceleration.ACCELERATING;
				break;
			case 'b':
				vca = VehicleEvent.VehicleControlAcceleration.BRAKING;
				break;
			case '-':
				vca = VehicleEvent.VehicleControlAcceleration.ROLLING;
				break;
			default:
				vca = null;
				// Does not happen...
				break;
			}
			VehicleEvent.VehicleControlTurning vct;
			switch( split[1].charAt(1) ) {
			case 'L':
				vct = VehicleEvent.VehicleControlTurning.HARDLEFT;
				break;
			case 'l':
				vct = VehicleEvent.VehicleControlTurning.LEFT;
				break;
			case '-':
				vct = VehicleEvent.VehicleControlTurning.STRAIGHT;
				break;
			case 'r':
				vct = VehicleEvent.VehicleControlTurning.RIGHT;
				break;
			case 'R':
				vct = VehicleEvent.VehicleControlTurning.HARDRIGHT;
				break;
			default:
				vct = null;
				// Does not happen...
				break;
			}
			int vehiclex = Integer.parseInt(split[2]);
			int vehicley = Integer.parseInt(split[3]);
			int vehicledir = Integer.parseInt(split[4]);
			int vehiclespeed = Integer.parseInt(split[5]);
			VehicleEvent ve = new VehicleEvent(timestamp,vca,vct,vehiclex,vehicley,vehicledir,vehiclespeed);
			synchronized(mvehicles) {
				for( VehicleHandler vh : mvehicles )
					vh.changed(ve);
			}
			int pos = 6;
			while( pos < split.length ) {
				char kind = split[pos++].charAt(0);
				switch( kind ) {
				case 'h':
					pos += 3;
					break;
				case 'b':
				case 'c':
					int x = Integer.parseInt(split[pos++]);
					int y = Integer.parseInt(split[pos++]);
					int r = Integer.parseInt(split[pos++]);
					Obstacle obs = new Obstacle(x,y,r);
					if( mdetobstacles.add(obs) ) {
						ObstacleEvent oe = new ObstacleEvent(timestamp,kind == 'b' ? MapStateEnum.BOULDER : MapStateEnum.CRATER,x,y,r);
						synchronized( mobstacles ) {
							for( ObstacleHandler oh : mobstacles )
								oh.recognized(oe);
						}
					}
					break;
				case 'm':
					x = Integer.parseInt(split[pos++]);
					y = Integer.parseInt(split[pos++]);
					int dir = Integer.parseInt(split[pos++]);
					int speed = Integer.parseInt(split[pos++]);
					MartianEvent me = new MartianEvent(timestamp,x,y,dir,speed);
					synchronized( mmartians ) {
						for( MartianHandler mh : mmartians )
							mh.recognized(me);
					}
					break;
				default:
					// Cannot happen...
					break;
				}
			}
			break;
		case 'B':
			timestamp = Integer.parseInt(split[0]);
			synchronized( mcrashs ) {
				for( CrashHandler ch : mcrashs )
					ch.crashed(timestamp);
			}
			break;
		case 'C':
			// Ignored
			break;
		case 'K':
			// Ignored
			break;
		case 'S':
			// Ignored
			break;
		case 'E':
			synchronized( mcleanups ) {
				for( CleanupBase cb : mcleanups )
					cb.doCleanup();
			}
			break;
		default:
			// Should not happen...
			break;
		}	
	}
	public void sendCommand(AccelerationCommand ac,TurningCommand tc) {
		byte[] buf = new byte[3];
		int pos = -1;
		switch( ac ) {
		case ACCELERATE:
			buf[++pos] = 'a';
			break;
		case BREAK:
			buf[++pos] = 'b';
			break;
		}
		switch( tc ) {
		case LEFT:
			buf[++pos] = 'l';
			break;
		case RIGHT:
			buf[++pos] = 'r';
			break;
		}
		buf[++pos] = ';';
		try {
			msockstream.write(buf,0,pos+1);
			msockstream.flush();
		} catch (IOException ex) {
			/* ignore */
			System.err.println("Caught IO Exception :( "+ex);
		}
	}
	
	public static void main(String[] args) {
		try {
			FileInputStream fis = new FileInputStream(args[0]);
			Connection con = new Connection(null,0);
			con.startReading(fis);
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (UnknownHostException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		
	}
}