package swtpl.marsrover.planner;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Point;
import java.util.ArrayList;
import java.util.Stack;

import utils.Metrics;
import swtpl.marsrover.gui.PaintObserver;
import swtpl.marsrover.gui.Painter;
import swtpl.marsrover.map.MapObject;
import swtpl.marsrover.net.Connection;
import swtpl.marsrover.net.MartianHandler;
import swtpl.marsrover.net.ObstacleEvent;

public class JohoPlanner extends Planner implements Painter {
	private final static int SAFEBASERADIUS = 4000;
	private final static int SAFEDIST = 1000;
	private final static int PASSDIST = 500;
	private final static int PLAN_AHEAD = 30000;
	private final static double STEP = 10;
	private final static double STEP_FACTOR = 1;

	class Destination {
		int x;
		int y;
		int radius;
		boolean goRight;
		
		public Destination(int x, int y, int radius, boolean goRight) {
			this.x = x;
			this.y = y;
			this.radius = radius;
			this.goRight = goRight;
		}
		
		public String toString() {
			return "Dest["+x+","+y+":"+radius+(goRight ? "R" : "L")+"]";
		}
		
	}
	
	private Destination homeBase = new Destination(0,0,0, false);
	private Destination nextDest;
	private ObstaclePath obspath;
	private int curGoal;
	private volatile boolean running;

	public JohoPlanner(Connection con, PaintObserver po) {
		super(con,po);
		obspath = new ObstaclePath();
		if (po != null) {
        po.addPainter(this);
	}
	}

	public void doCleanup() {
		super.doCleanup();
		running = false;
	}

	public void paintComponent(PaintObserver po) {
		if (!running)
			return;
		
		po.setColor(Color.MAGENTA);
		po.drawCircle(nextDest.x, nextDest.y, 2*nextDest.radius);
		checkObstacle(po);
	}
	
	public boolean checkObstacle(PaintObserver po) {
		double a = rover.getA();
		int angAdd;
		double s = rover.getSpeed();
		double d = rover.getDir();
		double x = rover.getX();
		double y = rover.getY();
		double step = STEP;
		Destination goal = this.nextDest;
		for (int t = 0; t < PLAN_AHEAD; t+= (int)step) {
			step *= STEP_FACTOR;
			
			int radius = goal.radius;
			double dist = Metrics.euklidDistSq((int)x, (int) y, goal.x, goal.y);
			double realDistSq = dist - ((double)radius*radius);
			long baseDist = Metrics.euklidDistSq((int)x, (int)y); 
			if (baseDist <= SAFEBASERADIUS*SAFEBASERADIUS) {
				return false;
			}
			if (baseDist <= Metrics.euklidDistSq(goal.x, goal.y)) {
				goal = homeBase;
			}		
			
			int destDir = Metrics.angleTowards((int) x, (int) y, goal.x, goal.y);
			int realDestDirDelta = Metrics.angle((int) Math.sqrt(realDistSq), radius);
			if (goal.goRight)
				destDir -= realDestDirDelta;
			else
				destDir += realDestDirDelta;
			int delta = Metrics.normalizedAngle((int)(destDir - d));

			boolean hardTurn = false;
			if (Math.abs(delta) < rc.getTenthDegreePerSecsTurn()/5) {
				angAdd = 0;
			} else if (Math.abs(delta) < 2*rc.getTenthDegreePerSecsHardTurn()/5) {
				angAdd = rc.getTenthDegreePerSecsTurn();
			} else {
				angAdd = rc.getTenthDegreePerSecsHardTurn();
				hardTurn = true;
			}
			if (delta < 0)
				angAdd = -angAdd;
			
			s += (int) (step * a - rover.getDrag() * step * s*s);
			if (s < 0)
				s = 0;
			x += Metrics.scaledCos(d, s*step/1000);
			y += Metrics.scaledSin(d, s*step/1000);
			d += step * angAdd / 1000.0;
			if (d > 1800.0)
				d -= 3600.0;
			else if (d < -1800.0)
				d += 3600.0;
			
			if (po != null)
				po.drawLine((int)x, (int)y, (int)x, (int)y);
			
			ObstacleEvent obstacle = quadtree.getField((int)x, (int)y);
			if (obstacle != null) {
				/* Found an obstacle on the way */
				boolean evadeRight;
                double diffDir=0;
                if (goal != nextDest) {
                	evadeRight = nextDest.goRight;
                } else if (hardTurn) {
					evadeRight = angAdd > 0;
                } else {
					diffDir = Metrics.normalizedAngle(Metrics.angleTowards((int)x, (int)y, 
							obstacle.getX(), obstacle.getY()) - (int)d);
					evadeRight = diffDir > 0;
				}
				nextDest = new Destination(obstacle.getX(), obstacle.getY(),
                        obstacle.getRadius()+SAFEDIST, evadeRight); 
                if (curGoal == -1)
                	curGoal = obspath.waypoints.size();
                obspath.add(curGoal, 
                		new ObstaclePath.Obstacle(obstacle.getX(), obstacle.getY(), 
                				obstacle.getRadius() + SAFEDIST, !evadeRight));
                rover.computeObstaclePath(obspath);

				return true;
			}
		}
		return false;
	}

    private boolean goalRelevant(Destination dest) {
    	return Metrics.euklidDistSq(rover.getX(), rover.getY()) > Metrics.euklidDistSq(dest.x, dest.y);
    }

	public void run() {
		while (true) {
	        System.err.println("Preparing Run");
			synchronized(rover) {
				try {
					while (!rover.positionKnown) {
						rover.wait();
					}
				} catch (InterruptedException ex) {
					System.err.println("Can't wait for position:"+ex);
					return;
				}
			}

			curGoal = -1;
			nextDest = homeBase;
			obspath.waypoints.clear();
			running = true;
			while (running) {
				try {
	               Thread.sleep(20);
				} catch (InterruptedException ex) {
					System.err.println("Can't wait for position:"+ex);
					return;
				}
				
				if (nextDest != homeBase && !goalRelevant(nextDest)) {
					nextDest = homeBase;
					curGoal = -1;
				}
				
				checkObstacle(null);

				double dist = Metrics.euklidDistSq(rover.getX(), rover.getY(), nextDest.x, nextDest.y);
				double realDist = dist - ((double)nextDest.radius*nextDest.radius);

				int destDir = Metrics.angleTowards(rover.getX(), rover.getY(), nextDest.x, nextDest.y);
				int realDestDirDelta = Metrics.angle((int) Math.sqrt((double)realDist), nextDest.radius);
				if (nextDest.goRight)
					destDir -= realDestDirDelta;
				else
					destDir += realDestDirDelta;
				
	    		rover.setDesiredDirection(destDir);
			}
		}
	}
}
