package swtpl.marsrover.planner;

import utils.Metrics;
import swtpl.marsrover.gui.PaintObserver;
import swtpl.marsrover.net.CleanupBase;
import swtpl.marsrover.net.RunContext;
import swtpl.marsrover.net.VehicleEvent;
import swtpl.marsrover.net.VehicleHandler;
import swtpl.marsrover.net.InitHandler;
import swtpl.marsrover.net.Connection;
import swtpl.marsrover.net.Connection.AccelerationCommand;
import swtpl.marsrover.net.Connection.TurningCommand;
import swtpl.marsrover.net.VehicleEvent.VehicleControlAcceleration;
import swtpl.marsrover.net.VehicleEvent.VehicleControlTurning;
import swtpl.marsrover.planner.ObstaclePath.Obstacle;

public class Rover implements InitHandler, VehicleHandler, CleanupBase {
	/** acceleration in m/s^2 */
	private double accel;
	/** brake acceleration (negative) m/s^2 */
	private double brake;
	/** drag in 1/µm */
	private double drag;

	boolean accelKnown;
	boolean brakeKnown;
	boolean positionKnown;

	public int currAcc, currTurn;

	Connection connection;

	private int x, y, speed, dir, time;
	private double a;
	VehicleControlAcceleration accstate;
	RunContext rc;
	
	public String toString() {
		return "[Rover at ("+x+","+y+"), speed: "+speed+", dir: "+dir+"]";
	}

	/* return the brake or accel factor
	 * depending on the current maneuver 
	 */
	public double getA() { return a; }
	
	public int getX() { return x; }
	public int getY() { return y; }
	public int getSpeed() { return speed; }
	public int getDir() { return dir; }
	
	public double getAccel() { return accel; }
	public double getBrake() { return brake; }
	public double getDrag() { return drag; }

	public Rover(Connection con) {
		con.registerInitHandler(this);
		con.registerVehicleHandler(this);
		connection = con;
	}

	public void initialized(RunContext rc) {
		this.rc = rc;
	}

	public void doCleanup() {
		currAcc = 0;
		currTurn = 0;
		positionKnown = false;
	}

	public void calibrateAccel(int s1, int deltaS, int deltaT) {
		int ms = rc.getMaxSpeedMMPerSecs();
		accel = deltaS / ((double) deltaT * (1 - (double) s1*s1/ms/ms));
		drag = accel / (double) ms/ms;
		brake = -accel;
	}

	public void calibrateBrake(int s1, int deltaS, int deltaT) {
		brake = (deltaS + drag*deltaT*(s1*s1)) / deltaT;
	}

	public void changed(VehicleEvent ve) {
		if (!accelKnown) {
			if (accstate == VehicleControlAcceleration.ACCELERATING
					&& ve.getAccelerationState() == VehicleControlAcceleration.ACCELERATING) {
				calibrateAccel(speed, ve.getSpeed()- speed, ve.getTimeStamp() - time);
				accelKnown = true;
			}	
		} else if (!brakeKnown 
				&& accstate == VehicleControlAcceleration.BRAKING) {
			if (ve.getAccelerationState() == VehicleControlAcceleration.BRAKING) {
				calibrateBrake(speed, ve.getSpeed()- speed, ve.getTimeStamp() - time);
				brakeKnown = true;
			}
		}		
		x = ve.getPosX();
		y = ve.getPosY();
		dir = ve.getDir();
		speed = ve.getSpeed();
		time = ve.getTimeStamp();
		accstate = ve.getAccelerationState();
		synchronized(this) {
			positionKnown = true;
			notify();
		}
		setState(ve.getAccelerationState(), ve.getTurningState());
	}
	
	/**
	 * Compute radius of HardTurn circles add a given speed.
	 * @param speed the speed
	 * @return the radius of circle.
	 */
	private int speedToRadius(int speed) {
		/* turnTime = 3600 / rc.getTenthDegreePerSecsHardTurn();
		 * diameter = speed * turnTime;
		 * radius = diameter / 2 / PI;
		 */
		return (int) (speed *  (1800.0 / Math.PI) / rc.getTenthDegreePerSecsHardTurn());
	}
	private int radiusToSpeed(int radius) {
		return (int) (radius * (Math.PI / 1800.0) * rc.getTenthDegreePerSecsHardTurn()); 
	}
	private double computeDist(double x, double y) {
		return Math.sqrt(x*x + y*y);
	}
	private double computeDir(double x, double y) {
		return Math.atan2(y, x);
	}
	
	public void computeObstaclePath(ObstaclePath path) {
		int lastX = x;
		int lastY = y;
		int lastApexDir = dir;
		boolean lastEvadeLeft = false;
		int maxRadius = speedToRadius(rc.getMaxSpeedMMPerSecs());
		for (int i = 0; i < path.waypoints.size(); i++) {
			Obstacle dest;
			Obstacle orig = path.waypoints.get(i);
			if (i+1 < path.waypoints.size()) {
				dest = path.waypoints.get(i+1);
			} else {
				dest = new Obstacle(0, 0, 0, !lastEvadeLeft);
			}
			double lastDir = computeDir(orig.x - lastX, orig.y - lastY);
			double nextDir = computeDir(dest.x - orig.x, dest.y - orig.y);
			double apexDir = (lastDir + nextDir) / 2.0;
			if (Math.abs(lastDir - nextDir) > Math.PI)
				apexDir = apexDir + (apexDir > 0 ? -Math.PI : Math.PI);
			orig.apexDir = (int) (apexDir * (1800.0 / Math.PI));
			orig.apexX = (int) (Math.sin(apexDir)*orig.radius);
			orig.apexY = -(int) (Math.cos(apexDir)*orig.radius);
			if (orig.evadeLeft) {
				orig.apexX = - orig.apexX;
				orig.apexY = - orig.apexY;
			}
			orig.apexX += orig.x;
			orig.apexY += orig.y;

			double dist = computeDist(orig.apexX - lastX, orig.apexY - lastY);
			lastX = orig.apexX;
			lastY = orig.apexY;
			lastDir = orig.apexDir;
			lastEvadeLeft = orig.evadeLeft;
		}
	}

	public void setDesiredTarget(int dx, int dy) {
		utils.Debug.println(toString()+dx+","+dy);
		setDesiredDirection( Metrics.angleTowards(x, y, dx, dy) );
	}

	public void setDesiredDirection(int absoluteDestDir) {
		setDesiredSpeedDir(rc.getMaxSpeedMMPerSecs(), absoluteDestDir);
	}
	
	public void setDesiredSpeedDir(int destSpeed, int absoluteDestDir) {
		int wantTurn, wantAccel;
		int delta = Metrics.normalizedAngle(absoluteDestDir-dir);
		//utils.Debug.println("[Rover] aiming delta : "+delta);
		if (speed <= destSpeed) {
			wantAccel = 1;
		} else if (speed - (int)(drag*speed*speed/400.0) <= destSpeed) {
			wantAccel = 0;
		} else
			wantAccel = -1;
		if (Math.abs(delta) < rc.getTenthDegreePerSecsTurn()/5)
			wantTurn = 0;
		else if (Math.abs(delta) < 2*rc.getTenthDegreePerSecsHardTurn()/5)
			wantTurn = 1;
		else
			wantTurn = 2;
		if (delta < 0)
			wantTurn = -wantTurn;

		setDesiredStearing(wantAccel, wantTurn);
	}

	public void setDesiredStearing(int acc, int turn) {
		a = (acc>0) ? accel : (acc<0) ? brake : 0; 
		while (currAcc != acc || currTurn != turn) {
			AccelerationCommand acccmd = AccelerationCommand.NONE;
			TurningCommand turncmd = TurningCommand.NONE;
			if (currAcc < acc) {
				acccmd = AccelerationCommand.ACCELERATE;
				currAcc++;
			} else if (currAcc > acc) {
				acccmd = AccelerationCommand.BREAK;
				currAcc--;
			}
			if (currTurn < turn) {
				turncmd = TurningCommand.LEFT;
				currTurn++;
			} else if (currTurn > turn) {
				turncmd = TurningCommand.RIGHT;
				currTurn--;
			}
			
			//utils.Debug.println("sending: "+acccmd + ", "+turncmd);
			connection.sendCommand(acccmd, turncmd);
		}
	}

	public void setState(VehicleControlAcceleration accelerationState,
			VehicleControlTurning turningState) {
		switch (accelerationState) {
		case BRAKING:
			currAcc = -1;
			break;
		case ROLLING:
			currAcc = 0;
			break;
		case ACCELERATING:
			currAcc = 1;
			break;
		}
		switch (turningState) {
		case HARDRIGHT:
			currTurn = -2;
			break;
		case RIGHT:
			currTurn = -1;
			break;
		case LEFT:
			currTurn = 1;
			break;
		case HARDLEFT:
			currTurn = 2;
			break;
		case STRAIGHT:
			currTurn = 0;
			break;
		}
	}
}
